from __future__ import annotations

import ast
from dataclasses import dataclass
from enum import Enum, auto
from pathlib import Path
from typing import Generator

from docstring_parser import Docstring, ParseError, parse


class CodeSegmentType(Enum):
    Function = auto()
    Class = auto()


@dataclass
class CodeSegment:
    """A class to hold a code segment and its corresponding docstring."""

    file_path: Path
    code_path: str
    lineno: int
    code: str
    doc: str
    parsed_doc: Docstring | None
    args: dict[str, str] | None = None
    returns: str | None = None
    seg_type: CodeSegmentType = CodeSegmentType.Function


class DocStatus(Enum):
    CORRECT = "CORRECT"
    INCORRECT = "INCORRECT"


@dataclass
class CodeSegmentReport:
    status: DocStatus
    issues: list[str]

    @staticmethod
    def correct() -> CodeSegmentReport:
        return CodeSegmentReport(status=DocStatus.CORRECT, issues=[])


def _parse_file(filepath: Path) -> Generator[CodeSegment]:
    code = filepath.read_text()
    tree = ast.parse(code)
    for node in ast.walk(tree):
        if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
            func_code = ast.get_source_segment(code, node)
            assert func_code is not None
            func_doc = ast.get_docstring(node) or ""
            lineno = getattr(node, "lineno", 1)
            func_name = node.name
            codepath = f"{filepath}:{lineno} {func_name}"
            parsed_doc = None
            try:
                parsed_doc = parse(func_doc)
            except ParseError:
                pass

            yield CodeSegment(
                file_path=filepath,
                code=func_code,
                doc=func_doc,
                lineno=lineno,
                code_path=f"{codepath}",
                parsed_doc=parsed_doc,
                args={
                    a.arg: ast.unparse(a.annotation)
                    for a in node.args.args
                    if a.annotation is not None
                }
                if node.args
                else None,
                returns=ast.unparse(node.returns) if node.returns else None,
                seg_type=CodeSegmentType.Function,
            )


def _parse_folder(folderpath: Path) -> Generator[CodeSegment]:
    for file in folderpath.rglob("*.py"):
        yield from _parse_file(file)


def code_docs_from_path(path: str | Path) -> Generator[CodeSegment]:
    path = path if isinstance(path, Path) else Path(path)
    if not path.exists():
        raise FileNotFoundError(f"Path {path} does not exist.")
    if not path.is_file() and not path.is_dir():
        raise ValueError(f"Path {path} is neither a file nor a directory.")

    if path.is_file():
        yield from _parse_file(path)
    else:
        yield from _parse_folder(path)



from pydolce.parser import CodeSegment


def check_duplicate_params(segment: CodeSegment) -> list[str] | None:
    if segment.parsed_doc is None:
        return None

    errors = []
    checked_params = set()
    for param in segment.parsed_doc.params:
        p_name = param.arg_name

        if p_name in checked_params:
            errors.append(
                f"Parameter '{p_name}' is documented multiple times in the docstring."
            )

        checked_params.add(p_name)

    return errors if errors else None



from pydolce.parser import CodeSegment


def check_params_type(segment: CodeSegment) -> list[str] | None:
    if segment.args is None or not segment.args or segment.parsed_doc is None:
        return None

    errors = []
    for param in segment.parsed_doc.params:
        p_name = param.arg_name
        p_type = param.type_name
        if p_type is None:
            # If the type is not documented, skip the check for this parameter
            # There is another rule to check for missing types
            continue

        if p_name not in segment.args:
            # Parameter documented but not in signature
            # There is another rule to check for missing parameters
            continue

        sig_type = segment.args.get(p_name)
        if sig_type is None:
            sig_type = "None"

        if str(sig_type).lower() != p_type.lower():
            errors.append(
                f"Parameter '{p_name}' has type '{sig_type}' in signature but '{p_type}' in docstring."
            )

    return errors if errors else None



from pydolce.parser import CodeSegment


def check_params_exist(segment: CodeSegment) -> list[str] | None:
    if segment.args is None or not segment.args or segment.parsed_doc is None:
        return None

    errors = []
    for param in segment.parsed_doc.params:
        p_name = param.arg_name

        if p_name not in segment.args:
            errors.append(f"Parameter '{p_name}' documented but not in signature.")

    return errors if errors else None



from dataclasses import dataclass
from typing import Callable

from pydolce.parser import CodeSegment, CodeSegmentType
from pydolce.rules.checkers.duplicate_params import check_duplicate_params
from pydolce.rules.checkers.param_description import check_params_descr
from pydolce.rules.checkers.param_exist import check_params_exist
from pydolce.rules.checkers.param_missing import check_params_missing
from pydolce.rules.checkers.param_type_missing import check_params_type_missing
from pydolce.rules.checkers.param_types import check_params_type
from pydolce.rules.checkers.return_missing import check_return_missing
from pydolce.rules.checkers.return_type import check_return_type
from pydolce.rules.checkers.valid_syntax import check_valid_syntax


@dataclass
class DocRule:
    ref: str
    description: str
    prompt: str | None = None
    check: Callable[[CodeSegment], list[str] | None] | None = None

    @property
    def level(self) -> int:
        return int(self.ref[3])

    @property
    def number(self) -> int:
        return int(self.ref[3:])

    @property
    def sub_number(self) -> int:
        return int(self.ref[4:])


ALL_RULES = [
    DocRule(
        ref="DOC101",
        description="Function is missing a docstring.",
        check=lambda segment: []
        if segment.seg_type == CodeSegmentType.Function and not segment.doc.strip()
        else None,
    ),
    DocRule(
        ref="DOC102",
        description="Class is missing a docstring.",
        check=lambda segment: []
        if segment.seg_type == CodeSegmentType.Class and not segment.doc.strip()
        else None,
    ),
    DocRule(
        ref="DOC103",
        description="Class docstring has invalid syntax.",
        check=check_valid_syntax,
    ),
    DocRule(
        ref="DOC201",
        description="Duplicate parameters in docstring.",
        check=check_duplicate_params,
    ),
    DocRule(
        ref="DOC202",
        description="Documented parameter does not exist",
        check=check_params_exist,
    ),
    DocRule(
        ref="DOC203",
        description="Missing parameter in documention",
        check=check_params_missing,
    ),
    DocRule(
        ref="DOC204",
        description="Parameter description is missing",
        check=check_params_descr,
    ),
    DocRule(
        ref="DOC205",
        description="Return missing from docstring",
        check=check_return_missing,
    ),
    DocRule(
        ref="DOC206",
        description="Parameter type missing",
        check=check_params_type_missing,
    ),
    DocRule(
        ref="DOC206",
        description="Invalid parameter type",
        check=check_params_type,
    ),
    DocRule(
        ref="DOC204",
        description="Invalid return type",
        check=check_return_type,
    ),
    DocRule(
        ref="DOC301",
        description="Docstring description contains spelling errors.",
        prompt="The docstring DESCRIPTION contains TYPOS. Exmaples: 'functon' instead of 'function', 'retrun' instead of 'return'. Report the specific typos. Scopes: [DESCRIPTION]",
    ),
    DocRule(
        ref="DOC302",
        description="Docstring parameter description contains spelling errors.",
        prompt="The description of some PARAMETERS contains TYPOS. Exmaples: 'functon' instead of 'function', 'retrun' instead of 'return'. Report the specific typos. Scopes: [PARAM_DESCRIPTION]",
    ),
    DocRule(
        ref="DOC303",
        description="Docstring return description contains spelling errors.",
        prompt="The description of the RETURN VALUE contains TYPOS. Exmaples: 'functon' instead of 'function', 'retrun' instead of 'return'. Report the specific typos. Scopes: [RETURN_DESCRIPTION]",
    ),
    DocRule(
        ref="DOC401",
        description="Docstring states the function does something that the code does not do.",
        prompt="The docstring summary does not match with the code summary. For example, the docstring says 'This function sends an email', but the code sends an SMS. Scopes: [DOCSTRING, CODE]",
    ),
    DocRule(
        ref="DOC402",
        description="Docstring omits a critical behavior that the code performs.",
        prompt="The code performs a CRITICAL behavior X, but the docstring does not mention this behavior. CRITICAL means heavy tasks. Non critical behavior may no be documented. Scopes: [DESCRIPTION, CODE]",
    ),
]

RULES_FROM_REF = {r.ref: r for r in ALL_RULES}


class RuleSet:
    def __init__(
        self, target: list[str] | None = None, disable: list[str] | None = None
    ):
        if target is None:
            target = list(RULES_FROM_REF.keys())
        if disable is None:
            disable = []

        self.rules = [
            rule for rule in ALL_RULES if rule.ref in target and rule.ref not in disable
        ]

        self.by_level: dict = {}
        for rule in self.rules:
            self.by_level.setdefault(rule.level, []).append(rule)

    def __hash__(self) -> int:
        return hash(tuple(sorted(r.ref for r in self.rules)))

    def contains_llm_rules(self) -> bool:
        return any(r.prompt is not None for r in self.rules)

    def llm_rules(self) -> list[DocRule]:
        return [r for r in self.rules if r.prompt is not None]

    def check(self, segment: CodeSegment) -> list[str]:
        issues = []
        for rule in self.rules:
            if rule.check is not None:
                errors = rule.check(segment)
                if errors is None:
                    continue
                if not errors:
                    issues.append(f"{rule.ref}: {rule.description}")
                    continue
                for error in errors:
                    issues.append(f"{rule.ref}: {rule.description} ({error})")
        return issues


DEFAULT_RULESET = RuleSet()



from pydolce.parser import CodeSegment


def check_params_type_missing(segment: CodeSegment) -> list[str] | None:
    if segment.args is None or not segment.args or segment.parsed_doc is None:
        return None

    errors = []
    for param in segment.parsed_doc.params:
        p_name = param.arg_name
        p_type = param.type_name
        if p_type is None:
            errors.append(f"Parameter '{p_name}' is missing a type in the docstring.")

    return errors if errors else None



from __future__ import annotations

import json
import re
from pathlib import Path
from typing import Counter

import rich

from pydolce.client import LLMClient, LLMConfig
from pydolce.config import DolceConfig
from pydolce.parser import (
    CodeSegment,
    CodeSegmentReport,
    DocStatus,
    code_docs_from_path,
)
from pydolce.rules.rules import RULES_FROM_REF, DocRule


def ruled_check_prompts(
    function_code: str,
    rules: list[DocRule],
) -> tuple[str, str]:
    """
    Create system and user prompts for the model to check if docstring follows the defined rules.

    This will NOT check parameters, returns values, or any other section but the
    main description of the docstring.

    This will NOT check for completeness, only for CRITICAL inconsistencies.

    Args:
        function_code: The Python function code to analyze
        existing_docstring: Current docstring (if any)

    Returns:
        Tuple of (system_prompt, user_prompt)
    """

    rules_str = "\n".join(
        f"- {r.ref}: {r.prompt if r.prompt else r.description}" for r in rules
    )
    system_prompt = """You are an expert Python docstring analyzer. Your task is to analyze if a Python function docstring follows a set of defined rules."""

    system_prompt += f"""
Analysis scopes:
- DOCSTRING: The entire docstring, including all sections.
- DESCRIPTION: The main description of the docstring.
- PARAM_DESCRIPTION: The description of each parameter in the docstring.
- RETURN_DESCRIPTION: The description of the return value in the docstring.
- DOC_PARAM: The entire parameter section of the docstring.
- PARAMS: The parameters in the function signature.
- CODE: The actual code of the function.

RULES TO CHECK:
{rules_str}
"""

    system_prompt += """
Go rule by rule, and check if the docstring violates any of them independently of the others. For each rule use only the scope information provided in the rule description to determine if the rule is violated or not.

EXACT OUTPUT FORMAT IN JSON:

```
{
    "status": "[CORRECT/INCORRECT]",
    "issues": [List of specific rules references (DOCXXX) that were violated. Empty if status is CORRECT.]
    "descr": [List of specific descriptions of the issues found, one per issue. No more than one sentence. Empty if status is CORRECT.]
}
```

VERY IMPORTANT: NEVER ADD ANY EXTRA COMENTARY OR DESCRIPTION. STICK TO THE EXACT OUTPUT FORMAT.
"""

    user_prompt = f"""
Check this code:
```python
{function_code.strip()}
```
"""
    return system_prompt, user_prompt


def _extract_json_object(text: str) -> str | None:
    start = text.find("{")
    if start == -1:
        return None

    brace_count = 0
    in_string = False
    escape_next = False

    for i, char in enumerate(text[start:], start):
        if escape_next:
            escape_next = False
            continue

        if char == "\\":
            escape_next = True
            continue

        if char == '"' and not escape_next:
            in_string = not in_string
            continue

        if not in_string:
            if char == "{":
                brace_count += 1
            elif char == "}":
                brace_count -= 1
                if brace_count == 0:
                    return text[start : i + 1]

    return None


def _print_summary(responses: list[CodeSegmentReport]) -> None:
    statuses_count = Counter(resp.status for resp in responses)
    rich.print("\n[bold]Summary:[/bold]")
    if DocStatus.CORRECT in statuses_count:
        rich.print(f"[green]✓ Correct: {statuses_count[DocStatus.CORRECT]}[/green]")
    if DocStatus.INCORRECT in statuses_count:
        rich.print(f"[red]✗ Incorrect: {statuses_count[DocStatus.INCORRECT]}[/red]")


def check_description(
    codeseg: CodeSegment, llm: LLMClient, rules: list[DocRule]
) -> CodeSegmentReport | None:
    sys_prompt, user_prompt = ruled_check_prompts(
        function_code=codeseg.code, rules=rules
    )
    response = llm.generate(
        prompt=user_prompt,
        system=sys_prompt,
    )

    json_resp_str = _extract_json_object(response)

    if json_resp_str is None:
        rich.print(
            "  [yellow]⚠ Invalid response from model. Ignoring function[/yellow]"
        )
        return None

    json_resp = json.loads(json_resp_str)

    if json_resp["status"] == DocStatus.CORRECT.value:
        return CodeSegmentReport.correct()

    if json_resp["issues"]:
        issues = []
        for i, issue in enumerate(json_resp["issues"]):
            ref_search = re.search(r"DOC\d{3}", issue)
            if ref_search is None:
                # Unknown issue format
                continue

            ref = ref_search[0]
            if ref not in RULES_FROM_REF:
                # Unknown rule reference
                continue
            rule_descr = RULES_FROM_REF[ref].description
            issue_descr = (
                json_resp["descr"][i]
                if "descr" in json_resp and len(json_resp["descr"]) > i
                else ""
            )

            issue_str = f"{ref}: {rule_descr}"
            if issue_descr:
                issue_str += f" ({issue_descr})"
            issues.append(issue_str)
        json_resp["issues"] = issues

    return CodeSegmentReport(
        status=DocStatus.INCORRECT,
        issues=json_resp["issues"],
    )


def check(path: str, config: DolceConfig) -> None:
    checkpath = Path(path)

    llm = None
    if config.rule_set.contains_llm_rules():
        llm = LLMClient(LLMConfig.from_dolce_config(config))
        if not llm.test_connection():
            rich.print("[red]✗ Connection failed[/red]")
            return

    reports: list[CodeSegmentReport] = []

    for pair in code_docs_from_path(checkpath):
        loc = f"[blue]{pair.code_path}[/blue]"
        rich.print(f"[  ...  ] [blue]{loc}[/blue]", end="\r")

        quick_issues = config.rule_set.check(pair)
        if quick_issues:
            rich.print(f"[red][ ERROR ][/red] {loc}")
            report = CodeSegmentReport(
                status=DocStatus.INCORRECT,
                issues=quick_issues,
            )
            for issue in report.issues:
                rich.print(f"[red]  - {issue}[/red]")
            reports.append(report)
            continue

        if llm is not None and pair.doc.strip():
            desc_report = check_description(pair, llm, config.rule_set.llm_rules())
            if desc_report is None:
                continue

            if desc_report.status != DocStatus.CORRECT:
                reports.append(desc_report)
                rich.print(f"[red][ ERROR ][/red] {loc}")
                for issue in desc_report.issues:
                    rich.print(f"[red]  - {issue}[/red]")
                continue

        reports.append(CodeSegmentReport.correct())
        rich.print(f"[green][  OK   ][/green] {loc}")

    _print_summary(reports)



from pydolce.parser import CodeSegment, CodeSegmentType


def check_return_type(segment: CodeSegment) -> list[str] | None:
    if segment.seg_type != CodeSegmentType.Function:
        return None
    if (
        segment.returns is None
        or not segment.returns
        or segment.parsed_doc is None
        or segment.parsed_doc.returns is None
    ):
        return None

    if segment.returns != segment.parsed_doc.returns.type_name:
        return [
            f"Return type is '{segment.returns}' in signature but '{segment.parsed_doc.returns.type_name}' in docstring."
        ]

    return None



from pydolce.parser import CodeSegment


def check_params_missing(segment: CodeSegment) -> list[str] | None:
    if segment.args is None or not segment.args or segment.parsed_doc is None:
        return None

    documented_params = {param.arg_name for param in segment.parsed_doc.params}
    errors = []
    for p_name in segment.args:
        if p_name not in documented_params:
            errors.append(f"Parameter '{p_name}' in signature but not documented.")

    return errors if errors else None



from pydolce.parser import CodeSegment


def check_params_descr(segment: CodeSegment) -> list[str] | None:
    if segment.parsed_doc is None:
        return None

    errors = []
    for param in segment.parsed_doc.params:
        p_descr = param.description
        if p_descr is None or not p_descr.strip():
            errors.append(f"Parameter '{param.arg_name}' is missing a description.")

    return errors if errors else None



import docstring_parser

from pydolce.parser import CodeSegment


def check_valid_syntax(segment: CodeSegment) -> list[str] | None:
    if segment.parsed_doc is not None:
        return None

    try:
        docstring_parser.parse(segment.doc)
    except docstring_parser.ParseError as e:
        return [f"{e}"]

    return None



from pydolce.parser import CodeSegment, CodeSegmentType


def check_params_descr(segment: CodeSegment) -> list[str] | None:
    if segment.seg_type != CodeSegmentType.Function:
        return None
    if (
        segment.parsed_doc is None
        or segment.parsed_doc.returns is None
        or (segment.returns is not None and segment.returns == "None")
    ):
        return None

    ret = segment.parsed_doc.returns
    if ret.description is None or not ret.description.strip():
        return []

    return None



from __future__ import annotations

import os
from dataclasses import dataclass
from pathlib import Path
from typing import Any

import toml

from pydolce.rules.rules import DEFAULT_RULESET, RuleSet

DEFAULT_EXCLUDES = [
    "__init__.py",
    "setup.py",
    "conftest.py",
    "tests/*",
    "test_*.py",
    "*_test.py",
    "*/tests/*",
    ".venv",
    ".git",
    "dist",
]


@dataclass
class DolceConfig:
    """Configuration for Dolce"""

    # General options
    target: list[str] | None = None  # Specific rules to target
    disable: list[str] | None = None  # Specific rules to disable
    rule_set: RuleSet = DEFAULT_RULESET

    exclude: list[str] | None = None

    # Signature options
    docsig_config: dict[str, Any] | None = None

    # LLM options
    provider: str = "ollama"
    url: str = "http://localhost:11434"
    model: str = "qwen3:8b"
    api_key: str | None = None
    temperature: float = 0.0
    max_tokens: int | None = 2000
    timeout: int = 120
    max_retries: int = 3
    retry_delay: float = 1.0

    def describe(self) -> str:
        """Return a string description of the configuration."""

        desc = f"Provider: {self.provider}\n"
        desc += f"URL: {self.url}\n"
        desc += f"Model: {self.model}\n"
        if self.api_key:
            desc += "API Key: [REDACTED]\n"
        else:
            desc += "API Key: Not Set\n"
        return desc

    @staticmethod
    def from_pyproject() -> DolceConfig:
        """Load configuration from pyproject.toml if available."""
        pyproject_path = Path("pyproject.toml")
        if not pyproject_path.exists():
            return DolceConfig()

        pyproject = toml.load(pyproject_path)
        tool = pyproject.get("tool", {})
        config = tool.get("dolce", {})

        if "target" in config or "disable" in config:
            config["rule_set"] = RuleSet(
                target=config.get("target", None), disable=config.get("disable", None)
            )
        else:
            config["rule_set"] = DEFAULT_RULESET

        if "exclude" not in config:
            config["exclude"] = DEFAULT_EXCLUDES

        api_key_env_var = config.get("api_key", None)
        config["api_key"] = (
            None if api_key_env_var is None else os.environ.get(api_key_env_var, None)
        )

        docsig = tool.get("docsig", {})
        docsig["no_ansi"] = True
        _docsig_config = {k.replace("-", "_"): v for k, v in docsig.items()}
        config["docsig_config"] = _docsig_config

        return DolceConfig(**config)

    def update(self, **kwargs: Any) -> None:
        """Update configuration attributes."""
        for key, value in kwargs.items():
            if value is not None and hasattr(self, key):
                setattr(self, key, value)



from pydolce.parser import CodeSegment, CodeSegmentType


def check_return_missing(segment: CodeSegment) -> list[str] | None:
    if segment.seg_type != CodeSegmentType.Function:
        return None

    if segment.parsed_doc is None:
        return None

    if segment.returns is not None and segment.returns == "None":
        return None

    return [] if not segment.parsed_doc.returns else None
